/*
Tims Lua Binding Generator is copyright 2008 T N Channon
Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. */

#define VERSION "Tim's Lua Binding Generator, Version 1.0a"
#define RIGHTS "Copyright 2008, T N Channon, under Lua standard licence."


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <errno.h>


#define NOERROR 0
#define ERROR !NOERROR

/* width used for writing out tables of constant data, just for keeping a sane line length */
#define WIDTH 16


/* list of standard headers for the C created source */
/* array items are printf ready */
static const char *includes[] =
    {   "#include <stdlib.h>\n",
        "#include <stdio.h>\n",
        "#include <string.h>\n",
        "#include \"lua.h\"\n",
        "#include \"lauxlib.h\"\n",
        NULL /* marker meaning this is end of array */
    };

/* structure which is data type for single linked list */
typedef struct llista {
    int kind;           /* kind of node data */
    int key;            /* node index, not actually used, meaning this was xth item added to list */
    char *field1;         /* pointer to string, dynamic allocated or NULL */
    char *field2;
    char *field3;
    struct llista *next; /* chain to next llist node or NULL */
}llist;


const char filesig[]={"Tim's_Lua_binder"};  /* if not found in config file is not valid file */


/* echo string data table to file */
static void write_strings(FILE *f, const char *data[])
{
    int i=0;
    while(data[i]) fprintf(f, "%s", data[i++]);
}

/* Simple linked list
* Needed to store text strings until written to output.
* Normal usage returns root of new dynamiacally allocated list on first call,
* save this in var and thereafter ignore return if you wish.
* First call like this
* mylist = list_add(kindof, text, (text|NULL), NULL);
* Further calls
* list_add(kindof, text, (text|NULL), mylist); */

static llist *list_add(int kind, char field1[], char field2[], char field3[], llist *list)
{
    static int key=0;
    llist *ll, *lold;

    ll=lold=list;
    if(ll == NULL)      /* NULL means make a new list */
    {
        ll=malloc(sizeof(llist));
        ll->next=NULL;
    }
    else        /* Is exist list and so traverse to last item for insert at end */
    {
        while(ll->next != NULL)
        {
            ll=ll->next;
        }
        ll->next=malloc(sizeof *ll); /* make new list item at end */
        ll=ll->next;
        ll->next=NULL;
    }
    ll->field1=ll->field2=ll->field3=NULL;
    ll->key=key, ll->kind=kind;         /* insert data into list item */
    key++;
    if(field1)
    {
        ll->field1=malloc(strlen(field1)+1);
        strcpy(ll->field1, field1);
    }
    if(field2)
    {
        ll->field2=malloc(strlen(field2)+1);
        strcpy(ll->field2, field2);
    }
    if(field3)
    {
        ll->field3=malloc(strlen(field3)+1);
        strcpy(ll->field3, field3);
    }    
    if(lold == NULL)    /* Return original list root which will be ignored or assigned, doesn't matter */
    {
        lold=ll;
    }
    return lold;
}

/* deallocate all strings and the list, caution with huge lists is recursive */
static llist *list_deallocate(llist *list)
{
        if(list)
        {
                if(list->next) list_deallocate(list->next);
                if(list->field3) free(list->field3);
                if(list->field2) free(list->field2);                
                if(list->field1) free(list->field1);
                free(list);
        }
        return NULL;
}


/* simple keyword matching and the sequence of keywords must match sequence in enums */
/* enumerated (yeah, in lower case). Is kinds of items used by the parser and generator */
enum {errorenum=0, sigenum, projectenum, callenum, scriptenum, runenum, includeenum, commentenum};

static int matchkeyw(char word[])
{
    static char *keyw[] =
    {
        ".",
        "Tim's_Lua_binder",
        "projectname",
        "call",
        "script",
        "run",
        "include",
        "#",
        NULL
    };
        
    int k=1;

    if(word[0] == '#')
    {
        k=commentenum;
    }
    else
    {
        while((keyw[k] != NULL) && (strcmp(word, keyw[k])!=0)) k++;

        if(keyw[k] == NULL) k=errorenum;
    }

    return k;
}

/* parses config file, returns linked list or NULL if a problem */
/* is a crude routine and could do with a proper parser, maybe one day */
/* general scheme is by line in config file
    <keyword> <item> <item> ||
    <keyword> <rest of line>        */
static llist *parseconfig(char filename[])
{
    FILE *f;
    llist *ll=NULL;

    int gotsig,gotproject,gotcall,gotscript;

    gotsig=gotproject=gotcall=gotscript=0;
    if(NULL == (f=fopen(filename, "rb")))
    {
        printf("Problem opening configuration file named %s\n", filename);
        ll=list_deallocate(ll);
    }
    else
    {
        char *buff;
        unsigned size;

        buff=NULL;
        /* ansi C deduction of file size */
        /* Routine fills buffer from file and passes to main routine */
        fseek(f, 0L, SEEK_END);
        size=ftell(f);
        fseek(f, 0L, SEEK_SET);
        buff = malloc(size+1);
        if(size != fread(buff, sizeof(char), size, f))
        {
            free(buff);
            printf("File access problem reading configuration file.\n");
        }
        else
        {
            int tok;
            char delim[] = {" \t\n\r"};     /* space tab or eol delimited */
            char delims[]= {"\n\r"};        /* whole line from where we are until end of line */
            char *p;
            char *p1;
            FILE *tempf;

            buff[size] = '\0';
            p = strtok( buff, delim);
            for(;;)
            {

                if(p == NULL)
                {
                    break;
                }
                tok=matchkeyw(p);
                switch(tok)
                {
                    case sigenum : gotsig++;
                        break;
                    case callenum:
                        p = strtok( NULL, delims);
                        ll=list_add(tok, p, NULL, NULL, ll);
                        break;
                    case projectenum:
                    case includeenum:
                    case commentenum:
                        p = strtok( NULL, delims);
                        ll=list_add(tok, p, NULL, NULL, ll);                            
                        break;
                    case scriptenum:
                    case runenum:
                        p = strtok( NULL, delim);
                        if(NULL == (tempf=fopen(p, "rb")))
                        {
                            printf("Can't open file %s\n", p);
                            tok=errorenum;
                        }
                        else
                        {
                            char *t1, *t2;
                            
                            fclose(tempf);
                            p1=malloc(strlen(p)+1);
                            strcpy(p1, p);
                            p = strtok( NULL, delim);

                            
                            /* de-dot name */
                            
                            t1=malloc(strlen(p)+1);
                            strcpy(t1, p);
                            while((t2=strchr(t1, '.')))
                            {
                                t2[0]='_';
                            }
                            ll=list_add(tok, p1, p, t1, ll);
                            free(t1); free(p1);
                            gotscript++;
                        }
                        break;
                    case errorenum:
                    default :
                        printf("Unknown reason for failure to do with configuration file.\nMaybe on this line %s\n", p);
                        tok=errorenum;
                        break;
                }
                if(tok==errorenum)
                {
                    ll=NULL;

                    break;
                }
                p = strtok( NULL, delim);
            }

        }
        if(buff != NULL)
        {
            free(buff);
        }

    }
    if(f != NULL)
    {
        fclose(f);
    }
    {
        char *p2;   /* what follows is ghastly, extract function name */
        char *buf;
        char *buf2;
        llist *tl;

        tl=ll;
        while(tl!=NULL)
        {
            if(tl->kind==callenum)
            {
                buf=malloc(strlen(tl->field1)+1);
                buf2=malloc(strlen(tl->field1)+1);
                strcpy(buf, tl->field1);  /* now mangle the call to extract what we need  */
                p2=strtok(buf, "(");    /* extract luaopen_mime_core                    */
                strcpy(buf, p2);        /* from eg. MIME_API int luaopen_mime_core(lua_State *L); */
                p2=strtok(buf, " ");    /* even if the user has put white space in front of (     */
                do
                {
                    strcpy(buf2, p2);
                    p2=strtok(NULL, " ");
                }
                while(p2);
                free(tl->field1);
                tl->field1=buf2;
                free(buf);
            }
            tl=tl->next;
        }
    }


    /* try to check for errors */
    if(gotsig&&gotproject&&gotcall&&gotscript==0)
    {
        printf("Missing item in configuration.\n");
        ll=list_deallocate(ll);
    }
    if(gotsig==0)
    {
        printf("Not found a valid configuration file signature.\n");
        ll=list_deallocate(ll);
    }
    if(gotproject > 1)
    {
        printf("Found more than one project name in configuration file.\n");
        ll=list_deallocate(ll);
    }
    return ll; /* save this return in a var, is root of linked list or NULL */
}

static int usage(void)
{
    printf("Tim's lua static binding generator, Usage: timslbg <name of config file>\n");
    return ERROR;
}

/* Put anti multiple includes interlock on file
*/

static void includeinterlock(FILE *f, const char name[])
{
    char *temp;
    int i, len;

    len=strlen(name)+1;
    temp=malloc(len);
    for(i=0;i<=len;i++)
    {
        if(name[i]=='.')
        {
            temp[i]='_';
        }
        else
        {
            temp[i]=toupper(name[i]);
        }
    }
    fprintf(f, "\n#ifndef _%s\n", temp);
    fprintf(f, "#define _%s\n", temp);
    fprintf(f, "#else\n#error Definitely a bad idea! Multiple includes of ths file.\n#endif\n\n");
    free(temp);
}

/* Timestamp file as comment, time value is time of first call and persistent through all later calls */
static void timestamp(FILE *outfile)
{
    /* note the persistent variables */
    static int need_to_be_initialised=1;
    time_t time_of_day;
    static char timebuf[32];    /* fixed buffer size okay with ansi c, persistent (global alloc data) */

    if(need_to_be_initialised)
    {
        time_of_day = time( NULL );
        strcpy(timebuf, ctime( &time_of_day ));
        timebuf[strlen(timebuf)-1]='\0'; /* lose lf from end of string otherwise output will give new line */
        need_to_be_initialised=0;
    }
    fprintf(outfile, "/* Timestamp( %s ) Automatically generated file.*/\n", timebuf);
    fprintf(outfile, "/* Generated by %s */\n\n", VERSION);
   
}

static char *makefilename(const char path[], const char name[], const char ext[])
{
    char *fn;

    int len=1;
    if(path) len+=strlen(path)+1;
    len+=strlen(name);
    if(ext) len+=strlen(ext)+1;
    fn=malloc(len);
    fn[0]='\0';
    if(path && (strlen(path)>0))
    {
        strcat(fn, path);
    }
    strcat(fn, name);
    if(ext)
    {
        strcat(fn, ext);
    }
    return fn;
}

/* This function does most of the work, called with a linked list filled from config file */
static int mainprog(llist *list)
{

    int ch; /* not char */
    char bindhname[256], bindcname[256], buf[256];
    FILE *bind_h, *bind_c;
    char *projectname;
    char *outpath = {""}; /* not active but used later on */
    llist *tl;
    int havepostrun=0;


    /* construct project names */
    tl=list;
    while((tl != NULL) && (tl->kind!=projectenum)) tl=tl->next;
    projectname=tl->field1;

    strcpy(buf, "imported_");
    strcat(buf, projectname);
    strcat(buf, "_script");

// bind files
    {
        const char bindprefix[] = {"bind_"};

        char *fullpath_h, *fullpath_c, *corename;
        char temp[256];

        strcpy(temp, bindprefix); strcat(temp, projectname);
        strcpy(bindhname, temp); strcat(bindhname, ".h");
        strcpy(bindcname, temp); strcat(bindcname, ".c");
        corename=makefilename(NULL,temp, NULL);
        fullpath_h=makefilename(outpath,corename,".h");
        fullpath_c=makefilename(outpath,corename,".c");


        bind_h=fopen(fullpath_h, "w");
        if(!bind_h)
        {
            int x;
            x=errno;
            exit(x);
        }
        bind_c=fopen(fullpath_c, "w");
        if(!bind_c)
        {
            int x;
            fclose(bind_h);
            x=errno;
            exit(x);
        }
    }

    timestamp(bind_h);
    timestamp(bind_c);
    includeinterlock(bind_c, bindcname);

    write_strings(bind_c, includes);
    fprintf(bind_c, "#include \"%s\"\n", bindhname);

    {
        llist *tl;

        tl=list;
         while(tl != NULL)
        {
            if(tl->kind==includeenum)
            {
                fprintf(bind_c, "%s\n", tl->field1);
            }
            tl=tl->next;
        }
    }

    fprintf(bind_c,"\n");



// import script files
    tl=list;
    while(tl != NULL)
    {
        FILE *f;
        unsigned length=0;

        if((tl->kind==scriptenum) || (tl->kind==runenum))
        {
            int cnt=6; /* offset start to make output neater */

            f=fopen(tl->field1, "rb");
            if(f == NULL)
            {
                printf("Script file %s not found.\n", tl->field1);
                return ERROR;
            }

            fprintf(bind_c,"static const char imported_%s_script[] = {", tl->field3);

            while ((ch = fgetc(f)) != EOF)
            {
                    fprintf(bind_c,"%#x,", ch);
                    cnt++; length++;
                    if(cnt > WIDTH)
                    {
                            cnt=0;
                            fprintf(bind_c, "\n\t");
                    }

            }
            fprintf(bind_c, "};\n\n");
            fclose(f);
            fprintf(bind_c, "/* Length in bytes of above constant data. */\n");
            fprintf(bind_c, "static const unsigned %s_script_length = %d;\n\n", tl->field3, length);
        }
        tl=tl->next;
    }

    {
        /* output source for function */
        
         const char *thecode[] = {
        "static int load_from_memory(lua_State *L, const char *data_array_name, const char *lua_handle, const unsigned data_length)\n",
        "{\n",
        "\tint status;\n\n",
        "\tstatus = luaL_loadbuffer(L, data_array_name, data_length, lua_handle) ||\n",
        "\t\tlua_pcall(L, 0, 0, 0);\n",
        "\tif(status)\n\t{\n",
        "\t\tfprintf(stdout, \"Loading %s ==> %s\\n\",lua_handle, lua_tostring(L,-1));\n",
        "\t}\n",
        "\treturn status;\n",
        "}\n\n",
        NULL
        };

        write_strings(bind_c, thecode);
    }

    fprintf(bind_c, "/* Following must be called immediately after lua open libs in your initialisation*/\n");
    fprintf(bind_c, "int  bind_%s_lib(lua_State *L)\n{\n", projectname);
        
// also write to header
    fprintf(bind_h, "int  bind_%s_lib(lua_State *L);\n\n", projectname);

    /* are there any post initialisation scripts to run? */
    { /* local block */

        tl=list;
        while(tl != NULL)
        {
            if(tl->kind==runenum)
            {
                havepostrun++;
            }
            tl=tl->next;
        }
    } /* end local block */

    /* if(havepostrun) */ /* decided to leave the following active regardless, makes dev of apps easier */
    {
        fprintf(bind_h, "int bind_%s_run(lua_State *L);\n\n", projectname);
    }   
    
    fprintf(bind_h, "/* End of file */\n");
    fclose(bind_h);  /* finished header file */

    /* now write out C library initialisation calls */
    { /* local block */

        tl=list;
        while(tl != NULL)
        {
            if(tl->kind==callenum)
            {
                fprintf(bind_c, "\t%s(L);\n ", tl->field1);
            }
            tl=tl->next;
        }
    } /* end local block */
    { /* start another, writes calls to load scripts at runtime */
        tl=list;
         while(tl != NULL)
        {
            if(tl->kind==scriptenum)
            {
                fprintf(bind_c, "\tload_from_memory(L, %s%s%s, \"%s\", %s_script_length);\n", "imported_", tl->field3, "_script", tl->field2, tl->field3);
            }
            tl=tl->next;
        }
    } /* end local block */
    fprintf(bind_c, "\treturn 1;\n}\n\n");

    /* Now write runtime Lua scripts if any */
/*    if(havepostrun) */ /* decided to leave active at all times to help ease development of apps */
    {
        fprintf(bind_c, "/* Following must be called when you are ready for the final script loads */\n");
        fprintf(bind_c, "int  bind_%s_run(lua_State *L)\n{\n", projectname);

        { /* local block */
    
            tl=list;
            while(tl != NULL)
            {
                if(tl->kind==runenum)
                {
                    fprintf(bind_c, "\tload_from_memory(L, %s%s%s, \"%s\", %s_script_length);\n", "imported_", tl->field3, "_script", tl->field2, tl->field3);
                }
                tl=tl->next;
            }
        } /* end local block */
        fprintf(bind_c, "\treturn 1;\n}\n\n"); 
    }
    
    fprintf(bind_c, "/* End of file */\n");
    fclose(bind_c);
    /* not really necessary to free memory, OS dumps usage on exit */
    free(projectname);
    return NOERROR;
}

int main(int argc, char *argv[])
{
    int ret;
    llist *ll;
    char buf[256];

    if(2 != argc)
    {
        ret=usage();
    }
    else
    {
        strcpy(buf, argv[1]);
        ll=parseconfig(buf);
        if(ll == NULL)
        {
            ret=usage();
        }
        else
        {

            ret=mainprog(ll);
        }
        ll=list_deallocate(ll);   
    }
    return ret;
}
